c=======================================================================
c     Adam Peplinski; 2012.03.23
c     Set of subroutines to solve eigenvalue problem with Arnoldi algorithm
c     using PARPACK.
c
c
c     Parameters used by this set of subroutines:
c     PARAM(69) - /= 0 if restart
c     PARAM(71) - checkpoint file number for restart (reading)
c     PARAM(72) - checkpoint file number for saving data
c     PARAM(108) - 0 if no arnoldi
c     PARAM(109) - frequency of calling arn_vsolve (number of time steps)
c     PARAM(110) - maximal number of arnoldi steps
c     PARAM(111) - size of Krylov space
c     PARAM(112) - number of calculated eigenvalues
c     PARAM(113) - PARPACK tolerance for Ritz values
c     PARAM(114) - 0 - direct mode; otherwise adjoint
c
c
C     Two possible modes are supported:
c     direct: A*x = lambda*x
c     inverse: A*x = lambda*M*x, M-mass matrix defining inner product
c     Inverse mode is preferred as it takes into account inner product
c     To define ARPACK mode: direct or inverse
c#define ARPACK_DIRECT
#undef ARPACK_DIRECT
c
c     ************************************************************
c     subroutine to initialise ARPACK
      subroutine arn_init

      INCLUDE 'SIZE'            ! NID, NDIM, NPERT
      INCLUDE 'TSTEP'           ! NSTEPS, LASTEP
      INCLUDE 'INPUT'           ! IFPERT, IFBASE, IFTRAN, PARAM, IF3D
      INCLUDE 'SOLN'            ! V?MASK, VXP, VYP, VZP
      INCLUDE 'MASS'            ! BM1
      INCLUDE 'ADJOINT'         ! IFADJ
      INCLUDE 'ARNOLDI'         !

!     ARPACK include file
      INCLUDE 'debug.h'

      integer i

c     global comunication in nekton
      integer NIDD,NPP,NEKCOMM,NEKGROUP,NEKREAL
      common /nekmpi/ NIDD,NPP,NEKCOMM,NEKGROUP,NEKREAL

c     set parameters
      IFARN = .TRUE.
      if (int(PARAM(108)).eq.0) IFARN = .FALSE.

      if (IFARN) then

c     timing
         ARTIME1=dnekclock()

c     check nekton parameters
         if (.not.IFTRAN) then
            if (NID.eq.0) write(6,*)
     $           'ERROR: arnoldi requres transient equations'
            call exitt
         endif

         if (NSTEPS.eq.0) then
            if (NID.eq.0) write(6,*)
     $           'ERROR: arnoldi requires NSTEPS>0'
            call exitt
         endif

         if (PARAM(12).ge.0) then
            if (NID.eq.0) write(6,*)
     $           'ERROR: I assume const. dt in arnoldi'
            call exitt
         endif

         if (.not.IFPERT) then
            if (NID.eq.0) write(6,*)
     $           'ERROR: arnoldi has to be run in perturbation mode'
            call exitt
         endif

         if (IFBASE) then
            if (NID.eq.0) write(6,*)
     $           'ERROR: arnoldi assumes constatnt base flow'
            call exitt
         endif

         if (NPERT.ne.1) then
            if (NID.eq.0) write(6,*)
     $           'ERROR: arnoldi requires NPERT=1'
            call exitt
         endif

c     do we start from checkpoint
         IFARNRST = .FALSE.
         if (int(PARAM(69)).eq.1) IFARNRST = .TRUE.

         NDIMA=int(PARAM(111))
         if (NDIMA.gt.LDIMA) then
            if (NID.eq.0) write(6,*) 'ERROR: NDIMA too big'
            call exitt
         endif

         NEGVA=int(PARAM(112))
         if (NEGVA.ge.(NDIMA/2)) then
            if (NID.eq.0) write(6,*) 'ERROR: NEGVA must be <NDIMA/2'
            call exitt
         endif

c     checkpointing data
         ARNISTART = int(PARAM(71))
         ARNISTOP  = int(PARAM(72))

c     direct or adjoint mode
         IFADJ = .TRUE.
         if (int(PARAM(114)).eq.0) IFADJ = .FALSE.

c     max steps in stepper and max cycles
         ARNSTEP = int(PARAM(109))
         ARNCMAX = int(PARAM(110))
c     set length of a single stepper stage
         ARNMSTEP = NSTEPS

         ARNISTEP = 0
         ARNTSTEP = 0
         ARNVSTEP = 0
         ARNCYC   = 0

c     for timing
         ARTIMET=0.0
         ARTIMES=0.0
         ARTIMESA=0.0

c     related to restart
         NPARP = 0
         NCARP = 0
         RNMARP= 0.0

c     initialize ARPACK parameters
#ifdef ARPACK_DIRECT
c     standard eigenvalue problem A*x = lambda*x
         BMATARP='I'
#else
c     generalized eigenvalue problem A*x = lambda*B*x
         BMATARP='G'
#endif
c     eigenvalues of largest magnitude
         WHICHARP='LM'
c     relative accuracy of the Ritz value
         TOLARP= PARAM(113)

         call izero(IPARP,11)
         call izero(IPNTARP,14)
c     exact shifts with respect to the current Hessenberg matrix
         IPARP(1)=1
c     maximum number of Arnoldi update iterations allowed
         IPARP(3)=ARNCMAX
#ifdef ARPACK_DIRECT
c     A*x = lambda*x
         IPARP(7)=1
#else
c     A*x = lambda*M*x, M symmetric positive definite; BMATARP='G'
         IPARP(7)=2
#endif
c     used size of WORKLA
         NWLARP = (3*NDIMA+6)*NDIMA

c     user supplied initial conditions
         INFARP=1

c     get eigenvectors
         RVARP=.true.
c     compute Ritz vectors
         HOWARP='A'
c     select should be specifird for HOWARP='S'

c     no shift
         SIGARP(1) = 0.0
         SIGARP(2) = 0.0
c     vector length
         NVECA  = NX1*NY1*NZ1*NELV
         NVECA2 = NVECA*NDIM

c     initialize arrays
         call rzero(WORKDA,WDDIMA)
         call rzero(WORKLA,WLDIMA)
         call rzero(WORKEA,WEDIMA)
         call rzero(VBASEA,LVA2*LDIMA)
         call rzero(RESIDA,LVA2)
         call rzero(DRIARP,LDIMA*4)

c     info level from ARPACK
         ndigit = -3
         logfil = 6
         mngets = 0
         mnaitr = 2
         mnapps = 0
         mnaupd = 2
         mnaup2 = 2
         mneupd = 0

c     PLACE FOR RESTART
         if (IFARNRST) then

c     read checkpoint
            call arn_rst_read

         else
c     if no restatrt fill RESIDA with initial conditions
c     V?MASK removes points at the wall and inflow
#ifdef ARPACK_DIRECT
c     A*x = lambda*x
            call col3(RESIDA(1),VXP,V1MASK,NVECA)
            call col3(RESIDA(1+NVECA),VYP,V2MASK,NVECA)
            if (IF3D) call col3(RESIDA(1+2*NVECA),VZP,V3MASK,NVECA)
#else
c     A*x = lambda*M*x
            call copy(RESIDA(1),VXP,NVECA)
            call copy(RESIDA(1+NVECA),VYP,NVECA)
            if (IF3D) call copy(RESIDA(1+2*NVECA),VZP,NVECA)
#endif

c     initialize rest of variables
c     firts call
            IDOARP=0
         endif

         call pdnaupd(NEKCOMM,IDOARP,BMATARP,NVECA2,WHICHARP,NEGVA,
     $        TOLARP,RESIDA,NDIMA,VBASEA,LVA2,IPARP,IPNTARP,WORKDA,
     $        WORKLA,NWLARP,INFARP,NPARP,RNMARP,NCARP)

 90      continue

c     error check
         if (INFARP.lt.0) then
            if (NID.eq.0) then
               write(6,*) 'ARNOLDI: '
               write(6,*) ' Error with _naupd, info = ', INFARP
               write(6,*) ' Check the documentation of _naupd.'
            endif
            call exitt
         endif

         if (IDOARP.eq.2) then
c     A*x = lambda*M*x
c     multiply by weights and masks
            call col3(WORKDA(IPNTARP(2)),BM1,V1MASK,NVECA)

            call col3(WORKDA(IPNTARP(2)+NVECA),BM1,V2MASK,NVECA)

            if (IF3D) call col3(WORKDA(IPNTARP(2)+2*NVECA),
     $           BM1,V3MASK,NVECA)


            call col2(WORKDA(IPNTARP(2)),WORKDA(IPNTARP(1)),NVECA2)

            call pdnaupd(NEKCOMM,IDOARP,BMATARP,NVECA2,WHICHARP,NEGVA,
     $           TOLARP,RESIDA, NDIMA,VBASEA,LVA2,IPARP,IPNTARP,WORKDA,
     $           WORKLA,NWLARP,INFARP,NPARP,RNMARP,NCARP)

c     error check
            if (INFARP.lt.0) then
               if (NID.eq.0) then
                  write(6,*) 'ARNOLDI: '
                  write(6,*) ' Error with _naupd, info = ', INFARP
                  write(6,*) ' Check the documentation of _naupd.'
               endif
               call exitt
            endif
         endif                  ! IDOARP

c     start stepper
         if (IDOARP.eq.-1.or.IDOARP.eq.1) then
c     move renormed data back to nekton
            call copy(VXP,WORKDA(IPNTARP(1)),NVECA)
            call copy(VYP,WORKDA(IPNTARP(1)+NVECA),NVECA)
            if (IF3D)
     $           call copy(VZP,WORKDA(IPNTARP(1)+2*NVECA),NVECA)

c     make sure the velocity field is continuous at element
c     faces and edges
            call opdssum(VXP,VYP,VZP)
            call opcolv (VXP,VYP,VZP,VMULT)
         elseif (IDOARP.eq.2) then
            goto 90
         endif

c     print info
         if (NID.eq.0) then
            write(6,*)
            write(6,*) 'ARPACK initialised'
            if (IFADJ) then
               write(6,*) 'ADJOINT mode'
            else
               write(6,*) 'DIRECT mode'
            endif
            write(6,*)
            write(6,*) 'Parameters:'
            write(6,'(A15,A1)') 'BMAT = ',BMATARP
            write(6,'(A15,A2)') 'WHICH = ',WHICHARP
            write(6,'(A15,G13.5)') 'TOL = ',TOLARP
            write(6,'(A15,I13)') 'NEV = ',NEGVA
            write(6,'(A15,I13)') 'NCV = ',NDIMA
            write(6,'(A15,I13)') 'IPARAM(1) = ',IPARP(1)
            write(6,'(A15,I13)') 'IPARAM(3) = ',IPARP(3)
            write(6,'(A15,I13)') 'IPARAM(7) = ',IPARP(7)
            write(6,'(A15,L)') 'RVEC = ',RVARP
            write(6,'(A15,A1)') 'HOWMNY = ',HOWARP
            if (HOWARP.eq.'S') then
               do i=1,LDIMA
                  write(6,100) i,SELARP(i)
               enddo
            endif
         endif

c     timing
         ARTIME2=dnekclock()
         ARTIMET=ARTIME2-ARTIME1

      endif

 100  FORMAT('  SELECT(',I2,') = ',L)

      return
      end
c     ************************************************************
c     subroutine to create Krylov space, get eigenvalues and restart
c     stepper phase
      subroutine arn_vsolve

      INCLUDE 'SIZE'            ! NID, NDIM, NPERT
      INCLUDE 'TSTEP'           ! ISTEP, NSTEPS, LASTEP
      INCLUDE 'INPUT'           ! IFHEAT, IF3D
      INCLUDE 'SOLN'            ! VXP, VYP, VZP, PRP, TP, VMULT, V?MASK
      INCLUDE 'MASS'            ! BM1
      INCLUDE 'ARNOLDI'         !

c     global comunication in nekton
      integer NIDD,NPP,NEKCOMM,NEKGROUP,NEKREAL
      common /nekmpi/ NIDD,NPP,NEKCOMM,NEKGROUP,NEKREAL

      if (IFARN) then

c     step counting
         ARNISTEP = ARNISTEP + 1
         ARNTSTEP = ARNTSTEP + 1
         if (mod(ARNISTEP,ARNSTEP).eq.0) then

c     stepper phase counting
            ARNISTEP = 0
            ARNVSTEP = ARNVSTEP +1

c     timing
            ARTIME1=dnekclock()
            ARTIMES=ARTIME1-ARTIME2
            ARTIME2=ARTIME1
            ARTIMESA=(ARTIMESA*(ARNVSTEP-1)+ARTIMES)/real(ARNVSTEP)

c     stamp output
            if (NID.eq.0) write(6,210) ARNVSTEP
 210        FORMAT('ARNOLDI: after',I5,' stepper phase')

c     fill work array with velocity 
c     V?MASK removes points at the boundary
#ifdef ARPACK_DIRECT
c     A*x = lambda*x
            call col3(WORKDA(IPNTARP(2)),VXP,V1MASK,NVECA)
            call col3(WORKDA(IPNTARP(2)+NVECA),VYP,V2MASK,NVECA)
            if (IF3D)
     $           call col3(WORKDA(IPNTARP(2)+2*NVECA),VZP,V3MASK,NVECA)
#else
c     A*x = lambda*M*x
            call copy(WORKDA(IPNTARP(2)),VXP,NVECA)
            call copy(WORKDA(IPNTARP(2)+NVECA),VYP,NVECA)
            if (IF3D)
     $           call copy(WORKDA(IPNTARP(2)+2*NVECA),VZP,NVECA)
#endif

            call pdnaupd(NEKCOMM,IDOARP,BMATARP,NVECA2,WHICHARP,NEGVA,
     $           TOLARP,RESIDA, NDIMA,VBASEA,LVA2,IPARP,IPNTARP,WORKDA,
     $           WORKLA,NWLARP,INFARP,NPARP,RNMARP,NCARP)

 230        continue

c     error check
            if (INFARP.lt.0) then
               if (NID.eq.0) then
                  write(6,*) 'ARNOLDI: '
                  write(6,*) ' Error with _naupd, info = ', INFARP
                  write(6,*) ' Check the documentation of _naupd.'
               endif
               call exitt
            endif

            if (IDOARP.eq.2) then
c     A*x = lambda*M*x
c     multiply by weights and masks
               call col3(WORKDA(IPNTARP(2)),BM1,V1MASK,NVECA)

               call col3(WORKDA(IPNTARP(2)+NVECA),BM1,V2MASK,NVECA)

               if (IF3D) call col3(WORKDA(IPNTARP(2)+2*NVECA),
     $              BM1,V3MASK,NVECA)


               call col2(WORKDA(IPNTARP(2)),WORKDA(IPNTARP(1)),NVECA2)

               call pdnaupd(NEKCOMM,IDOARP,BMATARP,NVECA2,WHICHARP,
     $              NEGVA,TOLARP,RESIDA,NDIMA,VBASEA,LVA2,IPARP,
     $              IPNTARP,WORKDA,WORKLA,NWLARP,INFARP,NPARP,RNMARP,
     $              NCARP)

c     error check
               if (INFARP.lt.0) then
                  if (NID.eq.0) then
                     write(6,*) 'ARNOLDI: '
                     write(6,*) ' Error with _naupd, info = ', INFARP
                     write(6,*) ' Check the documentation of _naupd.'
                  endif
                  call exitt
               endif
            endif               ! IDOARP

c     restart stepper
            if (IDOARP.eq.-1.or.IDOARP.eq.1) then
               if (NID.eq.0) write(6,*) 'ARNOLDI: restarting stepper'

c     put vector back as initial condition for next iteration
               call copy(VXP,WORKDA(IPNTARP(1)),NVECA)
               call copy(VYP,WORKDA(IPNTARP(1)+NVECA),NVECA)
               if (IF3D) call copy(VZP,WORKDA(IPNTARP(1)+2*NVECA),NVECA)

c     make sure the velocity field is continuous at element
c     faces and edges
               call opdssum(VXP,VYP,VZP)
               call opcolv (VXP,VYP,VZP,VMULT)

c     set time and iteration number !!!!!! possible problem
               TIME=0.0
               ISTEP=0          !!!!!!!!!!!!!!!!!!!!
c     zero presure and temperature
               call rzero(PRP,LVA1*LPERT)
               if (IFHEAT) call rzero(TP,LVA1*LPERT)

            elseif (IDOARP.eq.2) then
               goto 230
            endif

c     timing
            ARTIME3=dnekclock()
            ARTIMET=ARTIMET+ARTIME3-ARTIME1
         endif
      endif

      return
      end
c     ************************************************************
c     ARPACK postprocessing
      subroutine arn_esolve

      INCLUDE 'SIZE'            ! NID, NDIM, NPERT
      INCLUDE 'TSTEP'           ! ISTEP, DT, LASTEP
      INCLUDE 'SOLN'            ! VX, VY, VZ, VMULT, V?MASK
      INCLUDE 'INPUT'           ! IFPO,IFVO,IFTO,IFPSO(LDIMT1), IF3D
      INCLUDE 'ARNOLDI'         !

      integer i,n, ierror

      real dumm

      logical lifpo, lifvo, lifto, lifpso(LDIMT1)

c     global comunication in nekton
      integer NIDD,NPP,NEKCOMM,NEKGROUP,NEKREAL
      common /nekmpi/ NIDD,NPP,NEKCOMM,NEKGROUP,NEKREAL

      if (IFARN.and.IDOARP.eq.99) then

c     timing
         ARTIME1=dnekclock()

         if (NID.eq.0) write(6,405) IPARP(5)
 405     FORMAT('ARNOLDI: ',I4,' eigenvectors converged;
     $postprocessing')

         call pdneupd(NEKCOMM,RVARP,HOWARP,SELARP,DRIARP,DRIARP(1,2),
     $        VBASEA,LVA2,SIGARP(1),SIGARP(2),WORKEA,BMATARP,NVECA2,
     $        WHICHARP,NEGVA,TOLARP,RESIDA,NDIMA,VBASEA,LVA2,IPARP,
     $        IPNTARP,WORKDA,WORKLA,NWLARP,IERRARP)

         if (IERRARP.eq.0) then

            if (NID.eq.0) then
               write(6,*) 'ARNOLDI: writing eigenvalues'

c     open file 
               open (unit=55,file='eigenvalues.txt',err=406)
               write(55,*,err=406)
     $'          I;                re(RITZ);                im(RITZ);',
     $'                ln|RITZ|;               arg(RITZ)'

               dumm = DT*ARNSTEP
               dumm = 1.0/dumm
c     correct result; get eigenvalues of continuous operator
               do i=1,IPARP(5)
                  DRIARP(i,3) = log(sqrt(DRIARP(i,1)**2+
     $                 DRIARP(i,2)**2))*dumm
                  DRIARP(i,4) = atan2(DRIARP(i,2),DRIARP(i,1))*dumm
                  write(55,*,err=406) i,DRIARP(i,1),
     $                 DRIARP(i,2),DRIARP(i,3), DRIARP(i,4)
               enddo
               close(55)
            endif               ! NID

c     error check
            ierror=0
            goto 407
 406        ierror=1
 407        continue
            call err_chk(ierror,'Error writing eigenvalues.$')

c     write eigenvectors
            if (NID.eq.0) write(6,*) 'ARNOLDI: witing eigenvectors'
c     copy and set output parameters
            lifpo= IFPO
            IFPO = .false.
            lifvo= IFVO
            IFVO = .true.
            lifto= IFTO
            IFTO = .false.
            do i=1,LDIMT1
               lifpso(i)= IFPSO(i)
               IFPSO(i) = .false.
            enddo

c     we have to take into account sorrage of imaginary and real
c     parts of eigenvectors in arpack.
c     The complex Ritz vector associated with the Ritz value 
c     with positive imaginary part is stored in two consecutive 
c     columns.  The first column holds the real part of the Ritz 
c     vector and the second column holds the imaginary part.  The 
c     Ritz vector associated with the Ritz value with negative 
c     imaginary part is simply the complex conjugate of the Ritz
c     vector associated with the positive imaginary part.

            do i=1,IPARP(5)
c     copy eigenvectors to perturbation variables
               call copy(VXP,VBASEA(1,i),NVECA)
               call copy(VYP,VBASEA(1+NVECA,i),NVECA)
               if (IF3D) call copy(VZP,VBASEA(1+2*NVECA,i),NVECA)

               call outpost2(VXP,VYP,VZP,PRP,TP,0,'egv')
            enddo

c     put output variables back
            IFPO = lifpo
            IFVO = lifvo
            IFTO = lifto
            do i=1,LDIMT1
               IFPSO(i) = lifpso(i)
            enddo

         else                   ! IERRARP
            if (NID.eq.0) then
               write(6,*) 'ARNOLDI:'
               write(6,*) ' Error with _neupd, info = ', IERRARP
               write(6,*) ' Check the documentation of _naupd.'
            endif
            call exitt
         endif                  ! IERRARP

c     finish run
         lastep=1

c     timing
         ARTIME3=dnekclock()
         ARTIMET=ARTIMET+ARTIME3-ARTIME1
      endif

      return
      end
c     ************************************************************
c     subroutine to take care of savings for restart 
      subroutine arn_rst_save

      INCLUDE 'SIZE'            ! NID, NDIM, NPERT
      INCLUDE 'TSTEP'           ! ISTEP, NSTEPS, LASTEP
      INCLUDE 'ARNOLDI'         !

c     save checkpoint for IDOARP=-2
      if (IFARN.and.IDOARP.eq.-2) then

         if (NID.eq.0) then
            write(6,*) 'ARNOLDI: '
            write(6,*) 'IDOARP  = ', IDOARP
            write(6,*) 'Writing checkpoint'
         endif

c     save parameters and WORKLA; independent on processor; serial output
         call arn_write_par('ARP')

c     save big arrays; parallel output
         call mfo_arnv('ARV')

c     this is the last step
         lastep=1

      endif

      return
      end
c     ************************************************************
c     subroutine to read from checkpoints
      subroutine arn_rst_read

      INCLUDE 'SIZE'            ! NID, NDIM, NPERT
      INCLUDE 'TSTEP'           ! ISTEP, NSTEPS, LASTEP
      INCLUDE 'ARNOLDI'         !


      if (IFARN) then

         if (NID.eq.0) then
            write(6,*) 'ARNOLDI: '
            write(6,*) 'Reading checkpoint'
         endif
c     read parameters and WORKLA; independent on processor; serial input
         call arn_read_par('ARP')

c     read big arrays; parallel input
         call mfi_arnv('ARV')

      endif

      return
      end
c     ************************************************************
c     Adam Peplinski; 2012.01.27
c     subroutine to finalise arnoldi
      subroutine arn_end

      INCLUDE 'SIZE'            ! NID, NDIM, NPERT
      INCLUDE 'TSTEP'           ! ISTEP, NSTEPS, LASTEP
      INCLUDE 'ARNOLDI'         !

      if (IFARN.and.(LASTEP.eq.1)) then
         if (NID.eq.0) then
            write(6,*) ''
            write(6,*) 'ARNOLDI: finalize'
            write(6,*) '   Number of stepper cycles   ',ARNVSTEP
            write(6,*) '   Time spent in ARNOLDI      ',ARTIMET
            write(6,*) '   Average stepper phase time ',ARTIMESA
         endif
      endif

      return
      end
c-----------------------------------------------------------------------
c     Subroutines to manipulate files for arnoldi restart
c-----------------------------------------------------------------------
c     Subroutines to write/read procesor independent variables;
c     big arrays are saved/read by mfo_out_arnv/
c-----------------------------------------------------------------------
      subroutine arn_write_par(prefix)

      INCLUDE 'SIZE'
      INCLUDE 'INPUT'
      include 'RESTART'
      include 'TSTEP'
      INCLUDE 'ARNOLDI'

      character*3 prefix

      character*132 fname
      character*1   fname1(132)
      equivalence  (fname1,fname)

      character*6  str

      integer lwdsizo, k, ierr

      logical lifreguo

      call nekgsync()

c     copy and set output parameters
      lwdsizo= WDSIZO
      WDSIZO = 8

      lifreguo= IFREGUO
      IFREGUO = .false.

c     this is done by master node only
      if (NID.eq.0) then
c     create file name
         call mfo_filename(prefix,fname,SESSION,k)
         write(str,540) ARNISTOP
 540     format(i5.5)
         call chcopy(fname1(k),str,5)
        
c     open file; only serial
         call mbyte_open_serial(fname,NID,ierr)

         call mfo_arnp

c     close the file; only serial
         call byte_close(ierr)
      endif

c     put output variables back
      WDSIZO = lwdsizo

      IFREGUO = lifreguo

      return
      end
c-----------------------------------------------------------------------
      subroutine mfo_arnp

      include 'SIZE'
      include 'INPUT'
      include 'PARALLEL'
      include 'RESTART'
      include 'TSTEP'
      INCLUDE 'ADJOINT'         ! IFADJ
      INCLUDE 'ARNOLDI'

      character*16 hdr
      integer ahdsize
      parameter (ahdsize=16)

      integer ibsw_out, i, itmp(33), ierr

      real*4 test_pattern

      real*4 rtmp4(8), workla4(2*WLDIMA)
      real*8 rtmp8(4), workla8(WLDIMA)
      equivalence (rtmp4,rtmp8)
      equivalence (workla4,workla8)

c     write IDOARP and character varialbes 
      call blank(hdr,ahdsize)

      write(hdr,1) IDOARP,BMATARP,WHICHARP! 2+10=12
 1    format('#arp',1x,i2,1x,a1,1x,a2)

      ! if we want to switch the bytes for output
      ! switch it again because the hdr is in ASCII
      call get_bytesw_write(ibsw_out)
      if (ibsw_out.ne.0) call set_bytesw_write(0)  

      call byte_write(hdr,ahdsize/4,ierr)

c     write test pattern for byte swap
      test_pattern = 6.54321

      call byte_write(test_pattern,1,ierr)

c     collect and write integer varialbes
       itmp(1) = NVECA2
       itmp(2) = NEGVA
       itmp(3) = NDIMA
       itmp(4) = NWLARP
       itmp(5) = INFARP
       itmp(6) = NPARP
       itmp(7) = NCARP
       itmp(8) = ARNSTEP
       do i=1,11
          itmp(8+i) = IPARP(i)
       enddo
       do i=1,14
          itmp(19+i) = IPNTARP(i)
       enddo

       call byte_write(itmp,33,ierr)

c     collect and write real variables
       rtmp8(1) = TOLARP
       rtmp8(2) = RNMARP
       rtmp8(3) = DT
       if (IFADJ) then
          rtmp8(4) = 1.0
       else
          rtmp8(4) = 0.0
       endif

       call byte_write(rtmp4,8,ierr)

c     write WORKLA
       call copy(workla8,WORKLA,NWLARP)

       call byte_write(workla4,2*NWLARP,ierr)

      return
      end
c-----------------------------------------------------------------------
      subroutine arn_read_par(prefix)

      INCLUDE 'SIZE'
      INCLUDE 'INPUT'
      INCLUDE 'RESTART'
      INCLUDE 'TSTEP'
      INCLUDE 'PARALLEL'
      INCLUDE 'ADJOINT'         ! IFADJ
      INCLUDE 'ARNOLDI'

      character*3 prefix

      character*132 fname
      character*1   fname1(132)
      equivalence  (fname1,fname)

      character*6  str

      integer lwdsizo, k, ierr, error

      logical lifreguo

      call nekgsync()

c     copy and set output parameters
      lwdsizo= WDSIZO
      WDSIZO = 8

      lifreguo= IFREGUO
      IFREGUO = .false.

      error=0

c     this is done by master node only
      if (NID.eq.0) then
c     create file name
         call mfo_filename(prefix,fname,SESSION,k)
         write(str,540) ARNISTART
 540     format(i5.5)
         call chcopy(fname1(k),str,5)
        
c     open file; only serial
         call mbyte_open_serial(fname,NID, ierr)

c     read parameters
         call mfi_arnp

c     close the file; only serial
         call byte_close(ierr)

c     check and copy parameters

c     is it the same integration mode
         if (IFADJ.neqv.IFADJ0) then
            write(6,*) 'ERROR: arn_read_par, wrong IFADJ0; abort. '
            write(6,*) 'IFADJ0 = ', IFADJ0
            write(6,*) 'IFADJ  = ', IFADJ
            error=1
         endif

c     is it correct restart
         if (IDOARP0.ne.-2) then
            write(6,*) 'ERROR: arn_read_par, wrong IDOARP0; abort.'
            write(6,*) 'IDOARP0 = ', IDOARP0
            error=1
         endif

c     is it the same ARPACK mode
         if (BMATARP0.ne.BMATARP) then
            write(6,*) 'ERROR: arn_read_par, different ARPACK modes.'
            write(6,*) 'BMATARP0 = ', BMATARP0
            write(6,*) 'BMATARP  = ', BMATARP
            error=1
         endif

c     do we look for the same eigenvectors
         if (WHICHARP0.ne.WHICHARP) then
            write(6,*) 'ERROR: arn_read_par, different mode selection.'
            write(6,*) 'WHICHARP0 = ', WHICHARP0
            write(6,*) 'WHICHARP  = ', WHICHARP
            error=1
         endif

c     is the length of the vector the same
         if (NVECA20.ne.NVECA2) then
            write(6,*) 'ERROR: arn_read_par, different vector lenght;
     $abort.'
            write(6,*) 'NVECA20 = ', NVECA20
            write(6,*) 'NVECA2  = ', NVECA2
            error=1
         endif

c     what is the size of Krylov space
c     would it be possible to change this?; related NPARP, NWLARP,
c     IPNTARP
         if (NDIMA0.ne.NDIMA) then
            write(6,*) 'ERROR: arn_read_par, different Krylov space
     $size.'
            write(6,*) 'NDIMA0 = ', NDIMA0
            write(6,*) 'NDIMA  = ', NDIMA
            error=1
         endif

         if (NWLARP0.ne.NWLARP) then
            write(6,*) 'ERROR: arn_read_par, different size of work
     $array'
            write(6,*) 'NWLARP0 = ', NWLARP0
            write(6,*) 'NWLARP  = ', NWLARP
            error=1
         endif

c     stopping criterion 
         if (TOLARP0.ne.TOLARP) then
            write(6,*) 'WARNING: arn_read_par, different stopping
     $criterion'
            write(6,*) 'TOLARP0 = ', TOLARP0
            write(6,*) 'TOLARP  = ', TOLARP
         endif

c     number of eigenvalues
         if (NEGVA0.ne.NEGVA) then
            write(6,*) 'WARNING: arn_read_par, different number of 
     $eigenvalues'
            write(6,*) 'NEGVA0 = ', NEGVA0
            write(6,*) 'NEGVA  = ', NEGVA
         endif

c     stepper phase length
         if (DTARP0.ne.DT) then
            write(6,*) 'WARNING: arn_read_par, different timestep'
            write(6,*) 'DTARP0 = ', DTARP0
            write(6,*) 'DT     = ', DT
         endif

         if (ARNSTEP0.ne.ARNSTEP) then
            write(6,*) 'WARNING: arn_read_par, different number of 
     $steps in stepper phase'
            write(6,*) 'ARNSTEP0 = ', ARNSTEP0
            write(6,*) 'ARNSTEP  = ', ARNSTEP
         endif

c     check IPARP
         if (IPARP0(1).ne.IPARP(1)) then
            write(6,*) 'ERROR: arn_read_par, different shift in ARPACK'
            write(6,*) 'IPARP0(1) = ', IPARP0(1)
            write(6,*) 'IPARP(1)  = ', IPARP(1)
            error=1
         endif

         if (IPARP0(3).ne.IPARP(3)) then
            write(6,*) 'WARNING: arn_read_par, different cycle number'
            write(6,*) 'IPARP0(3) = ', IPARP0(3)
            write(6,*) 'IPARP(3)  = ', IPARP(3)
         endif

         if (IPARP0(7).ne.IPARP(7)) then
            write(6,*) 'ERROR: arn_read_par, different ARPACK modes'
            write(6,*) 'IPARP0(7) = ', IPARP0(7)
            write(6,*) 'IPARP(7)  = ', IPARP(7)
            error=1
         endif

c     copy rest of parameters
         NPARP = NPARP0
         NCARP = NCARP0
         INFARP= INFARP0
         RNMARP= RNMARP0
         do i=4,11
            IPARP(i) = IPARP0(i)
         enddo
         IPARP(2) = IPARP0(2)
         do i=1,14
            IPNTARP(i) = IPNTARP0(i)
         enddo
      endif                     ! NID

      call err_chk(error,'Error restarting arnoldi.$')

c      call nekgsync()


      IDOARP = -2
c     broadcast
      call bcast(NPARP,ISIZE)
      call bcast(NCARP,ISIZE)
      call bcast(INFARP,ISIZE)
      call bcast(IPARP,11*ISIZE)
      call bcast(IPNTARP,14*ISIZE)
      call bcast(RNMARP,WDSIZE)

      call bcast(WORKLA,NWLARP*WDSIZE)


c     put output variables back
      WDSIZO = lwdsizo

      IFREGUO = lifreguo

      return
      end
c-----------------------------------------------------------------------
      subroutine mfi_arnp

      include 'SIZE'
      include 'INPUT'
      include 'PARALLEL'
      include 'RESTART'
      include 'TSTEP'
      INCLUDE 'ARNOLDI'

      character*16 hdr
      character*4 dummy
      integer ahdsize
      parameter (ahdsize=16)

      integer ibsw_out, i, itmp(33), ierr

      real*4 test_pattern

      real*4 rtmp4(8), workla4(2*WLDIMA)
      real*8 rtmp8(4), workla8(WLDIMA)
      equivalence (rtmp4,rtmp8)
      equivalence (workla4,workla8)

      logical if_byte_swap_test, if_byte_sw_loc

c     read IDOARP and character varialbes 
      call blank(hdr,ahdsize)

      call byte_read(hdr,ahdsize/4,ierr)

      if (indx2(hdr,132,'#arp',4).eq.1) then
         read(hdr,*) dummy,IDOARP0,BMATARP0,WHICHARP0 ! 2+10=12
      else
         write(6,*) 'ERROR: mfi_arnp, wrong header; abort.'
         call exitt
      endif

c     read test pattern for byte swap
      call byte_read(test_pattern,1,ierr)
c     determine endianess
      if_byte_sw_loc = if_byte_swap_test(test_pattern,ierr) 

c     read integer varialbes
       call byte_read(itmp,33,ierr)
       if (if_byte_sw) call byte_reverse(itmp,33,ierr)

       NVECA20 = itmp(1)
       NEGVA0  = itmp(2)
       NDIMA0  = itmp(3)
       NWLARP0 = itmp(4)
       INFARP0 = itmp(5)
       NPARP0  = itmp(6)
       NCARP0  = itmp(7)
       ARNSTEP0 = itmp(8)
       do i=1,11
          IPARP0(i) = itmp(8+i)
       enddo
       do i=1,14
          IPNTARP0(i) = itmp(19+i)
       enddo

c     read real variables
       call byte_read(rtmp4,8,ierr)
       if (if_byte_sw) call byte_reverse(rtmp4,8,ierr)

       TOLARP0 = rtmp8(1)
       RNMARP0 = rtmp8(2)
       DTARP0  = rtmp8(3)
       if (rtmp8(4).eq.0.0) then
          IFADJ0 = .FALSE.
       else
          IFADJ0 = .TRUE.
       endif

c     read WORKLA
       if (NWLARP0.le.WLDIMA) then
          call byte_read(workla4,2*NWLARP0,ierr)
          if (if_byte_sw) call byte_reverse(workla4,2*NWLARP0,ierr)

          call copy(WORKLA,workla8,NWLARP0)
       else
          write(6,*) 'ARNOLDI: restart error'
          write(6,*) 'NWLARP0 = ',NWLARP0
          write(6,*) 'WLDIMA  = ',WLDIMA
          call exitt
       endif

      return
      end
c-----------------------------------------------------------------------
c     following subroutines are modiffications of 
c     
c     mfo_outfld
c     from prepost.f;
c     mfi
c     from ic.f
c-----------------------------------------------------------------------
      subroutine mfo_arnv(prefix)  ! muti-file output

      include 'SIZE'
      include 'TOTAL'
      include 'RESTART'
      INCLUDE 'ARNOLDI'

      integer*8 offs0,offs,nbyte,stride,strideB,nxyzo8

      integer lwdsizo, k, ierr

      character*3 prefix

      character*132  fname
      character*1    fname1(132)
      equivalence   (fname1,fname)

      character*6  str

      logical lifxyo, lifpo, lifvo, lifto, lifreguo, lifpso(LDIMT1)

 
      common /SCRUZ/  UR1(LXO*LXO*LXO*LELT)
     &              , UR2(LXO*LXO*LXO*LELT)
     &              , UR3(LXO*LXO*LXO*LELT)

      tiostart=dnekclock_sync()

c     copy and set output parameters
      lwdsizo= WDSIZO
      WDSIZO = 8

      lifreguo= IFREGUO
      IFREGUO = .false.
      lifxyo= IFXYO
      IFXYO = .false.
      lifpo= IFPO
      IFPO = .false.
      lifvo= IFVO
      IFVO = .true.
      lifto= IFTO
      IFTO = .false.
      do i=1,LDIMT1
         lifpso(i)= IFPSO(i)
         IFPSO(i) = .false.
      enddo

      nout = NELT
      NXO  = NX1
      NYO  = NY1
      NZO  = NZ1

c     set offset
      offs0 = iHeaderSize + 4 + isize*nelgt

      if (nid.eq.pid0) then         ! open files on i/o nodes
!         call mfo_open_files(prefix)
c     create file name
         call mfo_filename(prefix,fname,SESSION,k)
         write(str,5) ARNISTOP
 5       format(i5.5)
         call chcopy(fname1(k),str,5)
         call mbyte_open(fname,fid0,ierr)
c         write(6,*) nid,fid0,' FILE:',fname
      endif

      call mfo_write_hdr                     ! create element mapping +
                                             ! write hdr
      nxyzo8  = NXO*NYO*NZO
      strideB = nelB * nxyzo8*WDSIZO
      stride  = nelgt* nxyzo8*WDSIZO

      ioflds = 0
      ! dump all fields based on the t-mesh to avoid different
      ! topologies in the post-processor

c     resid array
      offs = offs0 + NDIM*strideB
      call byte_set_view(offs,ifh_mbyte)

      call copy(UR1,RESIDA(1),NVECA)
      call copy(UR2,RESIDA(1+NVECA),NVECA)
      if (IF3D) call copy(UR3,RESIDA(1+2*NVECA),NVECA)

      call mfo_outv(UR1,UR2,UR3,nout,NXO,NYO,NZO)
      ioflds = ioflds + NDIM

c     workd array
      do i=0,2
         offs = offs0 + ioflds*stride + NDIM*strideB
         call byte_set_view(offs,ifh_mbyte)

         call copy(UR1,WORKDA(1+NVECA2*i),NVECA)
         call copy(UR2,WORKDA(1+NVECA2*i+NVECA),NVECA)
         if (IF3D) call copy(UR3,WORKDA(1+NVECA2*i+2*NVECA),NVECA)

         call mfo_outv(UR1,UR2,UR3,nout,NXO,NYO,NZO)
         ioflds = ioflds + NDIM
      enddo

c     krylov space
      do i=1,NDIMA
         offs = offs0 + ioflds*stride + NDIM*strideB
         call byte_set_view(offs,ifh_mbyte)

         call copy(UR1,VBASEA(1,i),NVECA)
         call copy(UR2,VBASEA(1+NVECA,i),NVECA)
         if (IF3D) call copy(UR3,VBASEA(1+2*NVECA,i),NVECA)

         call mfo_outv(UR1,UR2,UR3,nout,NXO,NYO,NZO) 
         ioflds = ioflds + NDIM
      enddo

      dnbyte = 1.*ioflds*nout*WDSIZO*NXO*NYO*NZO

c     put output variables back
      WDSIZO = lwdsizo

      IFREGUO = lifreguo
      IFXYO = lifxyo
      IFPO = lifpo
      IFVO = lifvo
      IFTO = lifto
      do i=1,LDIMT1
         IFPSO(i) = lifpso(i)
      enddo

      if (NID.eq.PID0) 
#ifdef MPIIO
     &   call byte_close_mpi(ifh_mbyte,ierr)
#else
     &   call byte_close(ierr)
#endif

      tio = dnekclock_sync()-tiostart
      dnbyte = glsum(dnbyte,1)
      dnbyte = dnbyte + iHeaderSize + 4. +ISIZE*NELGT
      dnbyte = dnbyte/1024/1024
      if(nid.eq.0) write(6,7)  ISTEP,TIME,dnbyte,dnbyte/tio,
     &             NFILEO
    7 format(/,i9,1pe12.4,' done :: Write checkpoint',/,
     &       30X,'file size = ',3pG12.2,'MB',/,
     &       30X,'avg data-throughput = ',0pf7.1,'MB/s',/,
     &       30X,'io-nodes = ',i5,/)

      return
      end
c-----------------------------------------------------------------------
      subroutine mfi_arnv(prefix)
c
      include 'SIZE'
      include 'TOTAL'
      include 'RESTART'
      INCLUDE 'ARNOLDI'

      character*3 prefix

      character*132  fname
      character*1    fname1(132)
      equivalence   (fname1,fname)

      character*6  str

      character*132 hdr

      parameter (lwk = 7*lx1*ly1*lz1*lelt)
      common /scrns/ wk(lwk)
      integer e, k, iofldsr, ierr

      integer*8 offs0,offs,nbyte,stride,strideB,nxyzr8

      COMMON /scruz/ UR1 (LX1,LY1,LZ1,LELT)
     $ ,             UR2 (LX1,LY1,LZ1,LELT)
     $ ,             UR3 (LX1,LY1,LZ1,LELT)

      tiostart=dnekclock()

c     create file name
      call mfo_filename(prefix,fname,SESSION,k)
      write(str,5) ARNISTART
 5    format(i5.5)
      call chcopy(fname1(k),str,5)

      call mfi_prepare(fname)       ! determine reader nodes +
                                    ! read hdr + element mapping 

      offs0   = iHeadersize + 4 + ISIZE*NELGR
      nxyzr8  = NXR*NYR*NZR
      strideB = nelBr* nxyzr8*WDSIZR
      stride  = nelgr* nxyzr8*WDSIZR


c     read arrays
      iofldsr = 0
c     resid array
      offs = offs0 + NDIM*strideB
      call byte_set_view(offs,ifh_mbyte)
      call mfi_getv(UR1,UR2,UR3,wk,lwk,.false.)

      call copy(RESIDA(1),UR1,NVECA)
      call copy(RESIDA(1+NVECA),UR2,NVECA)
      if (IF3D) call copy(RESIDA(1+2*NVECA),UR3,NVECA)
      iofldsr = iofldsr + NDIM

c     workd array
      do i=0,2
         offs = offs0 + iofldsr*stride + NDIM*strideB
         call byte_set_view(offs,ifh_mbyte)
         call mfi_getv(UR1,UR2,UR3,wk,lwk,.false.)

         call copy(WORKDA(1+NVECA2*i),UR1,NVECA)
         call copy(WORKDA(1+NVECA2*i+NVECA),UR2,NVECA)
         if (IF3D) call copy(WORKDA(1+NVECA2*i+2*NVECA),UR3,NVECA)
         iofldsr = iofldsr + NDIM
      enddo

c     krylov space
      do i=1,NDIMA
         offs = offs0 + iofldsr*stride + NDIM*strideB
         call byte_set_view(offs,ifh_mbyte)
         call mfi_getv(UR1,UR2,UR3,wk,lwk,.false.)

         call copy(VBASEA(1,i),UR1,NVECA)
         call copy(VBASEA(1+NVECA,i),UR2,NVECA)
         if (IF3D) call copy(VBASEA(1+2*NVECA,i),UR3,NVECA)
         iofldsr = iofldsr + NDIM
      enddo

      nbyte = 0
      if(nid.eq.pid0r) nbyte = iofldsr*nelr*wdsizr*nxr*nyr*nzr

c     close files
#ifdef MPIIO
      if (nid.eq.pid0r) call byte_close_mpi(ifh_mbyte,ierr)
#else
      if (nid.eq.pid0r) call byte_close(ierr)
#endif
      call nekgsync
      tio = dnekclock()-tiostart

      dnbyte = nbyte
      nbyte = glsum(dnbyte,1)
      nbyte = nbyte + iHeaderSize + 4 + isize*nelgr

      if(nid.eq.0) write(6,7) istep,time,
     &             nbyte/tio/1024/1024/10,
     &             nfiler
    7 format(/,i9,1pe12.4,' done :: Read checkpoint data',/,
     &       30X,'avg data-throughput = ',f7.1,'MBps',/,
     &       30X,'io-nodes = ',i5,/)

      return
      end
c-----------------------------------------------------------------------
c     it is a copy of mbyte_open but without MPIIO part; I need it because
c     processor independent part is saved only by master.
c-----------------------------------------------------------------------
      subroutine mbyte_open_serial(hname,fid,ierr) ! open  blah000.fldnn
      include 'SIZE'
      include 'TSTEP'
      include 'RESTART'
 
      integer fid
      character*132 hname

      character*8  eight,fmt,s8
      save         eight
      data         eight / "????????" /

      character*132 fname
      character*1  fname1(132)
      equivalence (fname1,fname)

      integer      iname(20)
      equivalence (iname,fname)

      call izero  (iname,20)
      len = ltrunc(hname,132)
      call chcopy (fname,hname,len)

      do ipass=1,2      ! 2nd pass, in case 1 file/directory
         do k=8,1,-1
            i1 = indx1(fname,eight,k)
            if (i1.ne.0) then ! found k??? string
               write(fmt,1) k,k
    1          format('(i',i1,'.',i1,')')
               write(s8,fmt) fid
               call chcopy(fname1(i1),s8,k)
               goto 10
            endif
         enddo
   10    continue
      enddo
      
      call byte_open(fname,ierr)
      write(6,6) nid,istep,(fname1(k),k=1,len)
    6 format(2i8,' OPEN: ',132a1)

      return
      end
c-----------------------------------------------------------------------
