c-----------------------------------------------------------------------
c     VERSION FOR DNS AND LINEAR SIMULATIONS
c     This version supports only proper restart for DNS and for
c     perturbation mode with single(!!!!!) NPERT=1
c     perturbation and not advected basefield IFBASE=F. For DNS only
c     'rs8' files are written. For perturbation mode  files 'rs8'
c     include perturbation data and the basefield is saved in 'rb8'.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c
c     full-restart routines, called from userchk
c
c     saving and reading necessary files
c
c     Parameters used by this set of subroutines:
c     PARAM(69) - /= 0 if restart
c     PARAM(70) - checkpiont dump frequency (number of time steps)
c
c     PARAM(66) - write format
c     PARAM(67) - read format
c
c     In the case of multistep time-integration method one needs data 
c     from NBDINP timestep. I use standard full_restart and 
c     full_restart_save subroutines. There are four 'rs8...' reatart 
c     files saved in double precission. Only .f format is supported. 
c     In any case two sets of restart files (1-4;5-8) are created.
c
c     NOTICE!!!!
c     To make this varsion to work correctly
c     mod(total_number_of_steps,IOSTEPs).ge.NBDINP
c     Otherwise last checkpoint will not be complete.
c
c-----------------------------------------------------------------------
      subroutine my_full_restart

      include 'SIZE'            ! NID, NPERT
      include 'TSTEP'           ! IOSTEP, ISTEP
      include 'INPUT'           ! SESSION, IFREGUO, IFPERT, IFBASE
      include 'RESTART'         ! IFRIRO, NFILEO
      include 'PARALLEL'        ! ISIZE

      integer len, k, i, ierr
      integer i_set, nrsf, n_file
      save    i_set, n_file
      parameter  (nrsf=4)

      character*132 fname, bname
      character*1 fnam1(132)
      equivalence (fnam1,fname)

      character*3 prefix
      character*1 prefx1(3)
      equivalence (prefx1,prefix)

      character*80 s80(nrsf,15), rstfile
      save         s80, rstfile

      character*6  str

      character*17 kst
      save         kst
      data         kst / '0123456789abcdefx' /
      character*1  ks1(0:16)
      equivalence (ks1,kst)

      logical ifrestart, lifreguo
      save    ifrestart

      integer icalld
      save    icalld
      data    icalld  /0/

c     set some parameters
      lifreguo= IFREGUO
      IFREGUO = .false.

c     this is done only once
      if (icalld.eq.0) then
         icalld = icalld + 1

c     do we restart
         if (PARAM(69).eq.0.0) then
            ifrestart=.false.
         else
            ifrestart=.true.
         endif

c     check perturbation parameters
         if (IFPERT) then
            if (IFBASE.or.NPERT.gt.1.or.IFMHD) then
               if(NID.eq.0)
     $              write(6,*) 'FULL_RESTART: not supported mode'
               call exitt
            endif
         else                   ! IFPERT
            if (IFMHD) then
               if(NID.eq.0)
     $              write(6,*) 'FULL_RESTART: not supported mode'
               call exitt
            endif
         endif                  ! IFPERT

c     create rstfile name (SESSION.restart)
         call blank(fname,132)
         call blank(bname,132)
         call blank(rstfile,80)

         k = 1
         len = ltrunc(SESSION,132) !  Add SESSION
         call chcopy(fnam1(k),SESSION,len)
         k = k+len
         call chcopy(fnam1(k),'.restart',8)
         k = k+8
         call chcopy(rstfile,fname,k)

c     create names acording to mfo_open_files
         if (ifrestart) then 
c     create names of restart files
c     get set number from the file SESSION.restart
            ierr=0
            if(NID.eq.0) then
               open (unit=58,file=rstfile,err=50,status='old')
               read(58,*,err=50) len
               close(58)
               if(len.ne.0.and.len.ne.1) goto 50
               goto 51
 50            ierr=1
 51            continue
            endif
            call err_chk(ierr,'Error reading .restart file.$')

            call bcast(len ,ISIZE)
            i_set=len
            len = ltrunc(SESSION,132) !  Add SESSION
            call chcopy(bname,SESSION,len)

c     create file names
c     fill s80 array with 'rb8' and 'rs' file names
            if (IFPERT) then
c     assumes only single perturbation
               n_file = 2

c     prefix and name for basefield
               call chcopy(prefx1(1),'rb',2)
               len=min(17,2*nrsf)
               call chcopy(prefx1(3),ks1(len),1)
               call mfo_filename(prefix,fname,bname,k)

c     is fname too long?
               if ((k+5).gt.80) then
                  if(NID.eq.0) write(6,*)
     $                 'my_full_restart: too long file name'
                  call exitt
               endif

               do i=1,nrsf
                  call blank(s80(i,1),80)
c     this assume IFBASE=F
                  write(str,54) 1
c                     write(str,54) nrsf*i_set+i
 54               format(i5.5)
                  call chcopy(fnam1(k),str,5)
                  call chcopy(s80(i,1),fname,k+5)
c                     if (NID.eq.0) write(6,*) s80(i,1)
               enddo

c     create prefix and name for perturbation
               call chcopy(prefx1(1),'rs',2)
               len=min(17,2*nrsf)
               call chcopy(prefx1(3),ks1(len),1)
               call mfo_filename(prefix,fname,bname,k)

c     is fname too long?
               if ((k+5).gt.80) then
                  if(NID.eq.0) write(6,*)
     $                 'my_full_restart: too long file name'
                  call exitt
               endif

               do i=1,nrsf
                  call blank(s80(i,2),80)
                  write(str,54) nrsf*i_set+i
                  call chcopy(fnam1(k),str,5)
                  call chcopy(s80(i,2),fname,k+5)
c                     if (NID.eq.0) write(6,*) s80(i,2)
               enddo
            else                ! DNS
               n_file = 1

c     create prefix and name for DNS
               call chcopy(prefx1(1),'rs',2)
               len=min(17,2*nrsf)
               call chcopy(prefx1(3),ks1(len),1)
               call mfo_filename(prefix,fname,bname,k)

c     is fname too long?
               if ((k+5).gt.80) then
                  if(NID.eq.0) write(6,*)
     $                 'my_full_restart: too long file name'
                  call exitt
               endif

               do i=1,nrsf
                  call blank(s80(i,1),80)
                  write(str,54) nrsf*i_set+i
                  call chcopy(fnam1(k),str,5)
                  call chcopy(s80(i,1),fname,k+5)
c                     if (NID.eq.0) write(6,*) s80(i,1)
               enddo
            endif

         else                  ! ifrestart
!     set initial file set number
           i_set = 1
         endif                  ! ifrestart

         i_set = 1

      endif                     ! icalld

      if (ifrestart.and.(ISTEP.lt.nrsf))
     $     call full_restart_pert(s80,nrsf,n_file)

c      iosave = IOSTEP          ! Trigger save based on iostep
      iosave = int(PARAM(70))
      call full_restart_save_pert(iosave)

c     save file set in SESSION.restart
      ierr=0
      if(NID.eq.0) then
         iotest = 0
         if (ISTEP.gt.iosave/2  .and.
     $        mod(ISTEP+iosave-iotest,iosave).eq.(nrsf-1)) then

            if(i_set.eq.0) then
               i_set=1
            else
               i_set=0
            endif

            open (unit=58,file=rstfile,err=60)
            write(58,*,err=60) i_set
            close(58)
            goto 61
 60         ierr=1
 61         continue
         endif                  ! ISTEP
      endif                     ! NID
      call err_chk(ierr,'Error writing to .restart file.$')

c     put parameters back
      IFREGUO = lifreguo

      return
      end
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     VERSION FOR PERTURBATION MODE
c     following two subroutines are modiffications of 
c     full_restart_save
c     restart_save
c     from prepost.f.
c     This version supports only proper restart for single(!!!!!)
c     perturbation with not advected basefield IFBASE=F. In this case
c     files rs8 include perturbation and the basefield is saved in rb8.
c     The more general case (npert/=1 and IFBASE/=F) requeres increse
c     of file numbers to be written and e.g. hanges in restart_nfld.
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
      subroutine full_restart_save_pert(iosave)

      include 'SIZE'            ! NPERT, NID
      include 'INPUT'           ! IFPERT, IFBASE

      integer iosave,save_size,nfld_save

c     check perturbation parameters
      if ((IFPERT.and.(IFBASE.or.NPERT.gt.1)).or.IFMHD) then
         if(NID.eq.0) write(6,*) 'FULL_RESTART: not supported mode'
         call exitt
      endif

c     there is some problem with nfld_save; sometimes it would be good
c     to increase its value, but restart_nfld doesn't allow for that
      nfld_save=4  ! For full restart
      save_size=8  ! For full restart

      call restart_save_pert(iosave,save_size,nfld_save)

      return
      end
c-----------------------------------------------------------------------
      subroutine restart_save_pert(iosave,save_size,nfldi)

      integer iosave,save_size,nfldi


c     Save current fields for later restart.
c
c     Input arguments:
c
c       .iosave plays the usual triggering role, like iostep
c
c       .save_size = 8 ==> dbl. precision output
c
c       .nfldi is the number of rs files to save before overwriting
c

      include 'SIZE'
      include 'TOTAL'
      include 'RESTART'

      character*3 prefix, bprefix

      character*17 kst
      save         kst
      data         kst / '0123456789abcdefx' /
      character*1  ks1(0:16)
      equivalence (ks1,kst)

      logical if_full_pres_tmp

      integer icalld
      save    icalld
      data    icalld  /0/

      iosav = iosave

      if (iosav.eq.0) iosav = iostep
      if (iosav.eq.0) return

      iotest = 0
c     if (iosav.eq.iostep) iotest = 1  ! currently spoiled because of 
c                                      ! incompatible format of .fld
c                                      ! and multi-file i/o;  the latter
c                                      ! is the only form used for restart

      nfld  = nfldi*2
      nfld2 = nfld/2
      mfld  = min(17,nfld)
c     this is not supported; only second order time integration for MHD?
c     Why only 2 files per set not 3?
c      if (ifmhd) nfld2 = nfld/4

      i2 = iosav/2
      m1 = istep+iosav-iotest
      mt = mod(istep+iosav-iotest,iosav)
      prefix = '   '
      bprefix = '   '

      if (istep.gt.iosav/2  .and.
     $   mod(istep+iosav-iotest,iosav).lt.nfld2) then ! save
         write(prefix,3) ks1(mfld)
 3       format('rs',a1)
         write(bprefix,4) ks1(mfld)
 4       format('rb',a1)

         iwdsizo = wdsizo
         wdsizo  = save_size
         p66 = param(66)
         param(66) = 6                       ! force multi-file out

         npscal1 = npscal+1
         if (.not.ifheat) npscal1 = 0

         if_full_pres_tmp = if_full_pres     


         if (save_size.eq.8) if_full_pres = .true. !Preserve mesh 2 pressure

         if (IFPERT) then

c     save basefiled
c     do this only once
            if (icalld.eq.0) then
               icalld=1
               call outpost2(vx,vy,vz,pr,t,npscal1,bprefix)
            endif

c     save perturbation
            call outpost2(vxp(1,1),vyp(1,1),vzp(1,1),prp(1,1),tp(1,1,1)
     $           ,npscal1,prefix) ! perturbation
         else                   ! DNS
            call outpost2(vx,vy,vz,pr,t,npscal1,prefix) ! basefield
         endif

         wdsizo    = iwdsizo  ! Restore output parameters

         param(66) = p66
         if_full_pres = if_full_pres_tmp

      endif

c     if (nid.eq.0) write(6,8) istep,prefix,nfld,nfld2,i2,m1,mt
c  8  format(i8,' prefix ',a3,5i5)

      if_full_pres = .false.
      return
      end
c-----------------------------------------------------------------------
c-----------------------------------------------------------------------
c     VERSION FOR PERTURBATION MODE
c     following two subroutines are modiffications of 
c     full_restart
c     restart
c     mfi
c     map_pm1_to_pr
c     from ic.f.
c-----------------------------------------------------------------------
      subroutine full_restart_pert(s80,n_restart,n_file)

      include 'SIZE'
      include 'TOTAL'

      character*80 s80(4,15)

      ifile = istep+1  ! istep=0,1,...

      if (ifile.le.n_restart) then
         p67 = param(67)
         param(67) = 6.00
         do i=1,n_file
            call chcopy (initc(i),s80(ifile,i),80)
         enddo
         call bcast  (initc,132*n_file)

         if (IFPERT) then       ! perturbation
c     for IFBASE=F basefield loaded during istep=0 only
c     perturbation from 'rs' files in all steps
            if (IFBASE) then
               call restart_pert(n_file,1)
            elseif (ifile.eq.1) then
               call restart_pert(n_file,1)
            else
               call restart_pert(n_file,2)
            endif
         elseif (IFMHD) then    ! MHD
            call restart(2)
         else                   ! DNS
            call restart(1)
         endif                  ! IFPERT

         param(67)=p67
      endif
   
      return
      end
c-----------------------------------------------------------------------
c     this version supports .f (param(67).eq.6.0) format only
c-----------------------------------------------------------------------
      subroutine restart_pert(nfiles,nfilstart)
C----------------------------------------------------------------------
C
C     (1) Open restart file(s)
C     (2) Check previous spatial discretization 
C     (3) Map (K1,N1) => (K2,N2) if necessary
C
C     nfiles > 1 has several implications:
C
C     i.   For std. run, data is taken from last file in list, unless
C          explicitly specified in argument list of filename
C
C     ii.  For MHD and perturbation cases, 1st file is for U,P,T;
C          subsequent files are for B-field or perturbation fields
C
C
C     Adam Peplinski
C     There is additional input parameter specifying starting number
C     for file loop. It can be useful for perturbation mode with not
C     advected baseflow.
C----------------------------------------------------------------------
      INCLUDE 'SIZE'            ! NID
      INCLUDE 'TSTEP'           ! TIME
      INCLUDE 'INPUT'           ! PARAM
      INCLUDE 'RESTART'

      integer nfiles, nfilstart
      logical ifbasefl

      integer ifile,ndump

      character*132 fname
c
      if(nfiles.lt.1) return
      if(NID.eq.0) write(6,*) 'Reading checkpoint data'

c use new reader (only binary support)
      if (PARAM(67).eq.6.0) then
         do ifile=nfilstart,nfiles
            call sioflag(ndumps,fname,initc(ifile))
            call mfi_pert(fname,ifile)
         enddo
         call setup_convect(3)
         if (nid.ne.0) time=0
         time = glmax(time,1) ! Sync time across processors
         return
      else
         if (NID.eq.0) write(6,*) 'RESTART_PERT supporst .f only'
         call exitt
      endif

      return
      end
C----------------------------------------------------------------------
      subroutine mfi_pert(fname,ifile)
c
c     (1) Open restart file(s)
c     (2) Check previous spatial discretization 
c     (3) Map (K1,N1) => (K2,N2) if necessary
c
c     nfiles > 1 has several implications:
c
c     i.   For std. run, data is taken from last file in list, unless
c          explicitly specified in argument list of filename
c
c     ii.  For MHD and perturbation cases, 1st file is for U,P,T;
c          subsequent files are for B-field or perturbation fields
c
c
      include 'SIZE'
      include 'TOTAL'
      include 'RESTART'
      character*132 hdr
      character*132  fname

      logical if_full_pres_tmp

      parameter (lwk = 7*lx1*ly1*lz1*lelt)
      common /scrns/ wk(lwk)
      common /scrcg/ pm1(lx1*ly1*lz1,lelv)
      integer e, ierr

      integer*8 offs0,offs,nbyte,stride,strideB,nxyzr8

      tiostart=dnekclock()

      call mfi_prepare(fname)       ! determine reader nodes +
                                    ! read hdr + element mapping 

      offs0   = iHeadersize + 4 + isize*nelgr
      nxyzr8  = nxr*nyr*nzr
      strideB = nelBr* nxyzr8*wdsizr
      stride  = nelgr* nxyzr8*wdsizr

      if_full_pres_tmp = if_full_pres
      if (wdsizr.eq.8) if_full_pres = .true. !Preserve mesh 2 pressure

      iofldsr = 0
      if (ifgetxr) then      ! if available
         offs = offs0 + ndim*strideB
         call byte_set_view(offs,ifh_mbyte)
         if (ifgetx) then
c            if(nid.eq.0) write(6,*) 'Reading mesh'
            call mfi_getv(xm1,ym1,zm1,wk,lwk,.false.)
         else                ! skip the data
            call mfi_getv(xm1,ym1,zm1,wk,lwk,.true.)
         endif
         iofldsr = iofldsr + ndim
      endif

      if (ifgetur) then
         offs = offs0 + iofldsr*stride + ndim*strideB
         call byte_set_view(offs,ifh_mbyte)
         if (ifgetu) then
c     MHD
            if (ifmhd.and.ifile.eq.2) then
c               if(nid.eq.0) write(6,*) 'Reading B field'
               call mfi_getv(bx,by,bz,wk,lwk,.false.)
c     perturbation mode
c     importatn assumption; there are no MHD perturbation
            elseif (ifpert.and.ifile.ge.2) then
               j=ifile-1  ! pointer to perturbation field
               call mfi_getv(vxp(1,j),vyp(1,j),vzp(1,j),wk,lwk,.false.)
            else
c               if(nid.eq.0) write(6,*) 'Reading velocity field'
               call mfi_getv(vx,vy,vz,wk,lwk,.false.)
            endif
         else
            call mfi_getv(vx,vy,vz,wk,lwk,.true.)
         endif
         iofldsr = iofldsr + ndim
      endif

      if (ifgetpr) then
         offs = offs0 + iofldsr*stride + strideB
         call byte_set_view(offs,ifh_mbyte)
         if (ifgetp) then
c           if(nid.eq.0) write(6,*) 'Reading pressure field'
            call mfi_gets(pm1,wk,lwk,.false.)
         else
            call mfi_gets(pm1,wk,lwk,.true.)
         endif
         iofldsr = iofldsr + 1
      endif

      if (ifgettr) then
         offs = offs0 + iofldsr*stride + strideB
         call byte_set_view(offs,ifh_mbyte)
         if (ifgett) then
c     perturbation mode
            if (ifpert.and.ifile.ge.2) then
               j=ifile-1  ! pointer to perturbation field
               call mfi_gets(tp(1,1,j),wk,lwk,.false.)
            else
c               if(nid.eq.0) write(6,*) 'Reading temperature field'
               call mfi_gets(t,wk,lwk,.false.)
            endif
         else
            call mfi_gets(t,wk,lwk,.true.)
         endif
         iofldsr = iofldsr + 1
      endif

      do k=1,ldimt-1
         if (ifgtpsr(k)) then
            offs = offs0 + iofldsr*stride + strideB
            call byte_set_view(offs,ifh_mbyte)
            if (ifgtps(k)) then
c               if(nid.eq.0) write(6,'(A,I2,A)') ' Reading ps',k,' field'
c     perturbation mode
               if (ifpert.and.ifile.ge.2) then
                  j=ifile-1     ! pointer to perturbation field
                  call mfi_gets(tp(1,k+1,j),wk,lwk,.false.)
               else
                  call mfi_gets(t(1,1,1,1,k+1),wk,lwk,.false.)
               endif
            else
               call mfi_gets(t(1,1,1,1,k+1),wk,lwk,.true.)
            endif
            iofldsr = iofldsr + 1
         endif
      enddo
      nbyte = 0
      if(nid.eq.pid0r) nbyte = iofldsr*nelr*wdsizr*nxr*nyr*nzr

      if (ifgtim) time = timer

#ifdef MPIIO
      if (nid.eq.pid0r) call byte_close_mpi(ifh_mbyte,ierr)
#else
      if (nid.eq.pid0r) call byte_close(ierr)
#endif
      call nekgsync
      tio = dnekclock()-tiostart

      dnbyte = nbyte
      nbyte = glsum(dnbyte,1)
      nbyte = nbyte + iHeaderSize + 4 + isize*nelgr

      if(nid.eq.0) write(6,7) istep,time,
     &             nbyte/tio/1024/1024/10,
     &             nfiler
    7 format(/,i9,1pe12.4,' done :: Read checkpoint data',/,
     &       30X,'avg data-throughput = ',f7.1,'MBps',/,
     &       30X,'io-nodes = ',i5,/)

c     Adam Peplinski;  axis_interp_ic not modified
      if (ifaxis) call axis_interp_ic(pm1)      ! Interpolate to axi mesh
      if (ifgetp) call map_pm1_to_pr_pert(pm1,ifile) ! Interpolate pressure

      if_full_pres = if_full_pres_tmp

      return
      end
c-----------------------------------------------------------------------
      subroutine map_pm1_to_pr_pert(pm1,ifile)

      include 'SIZE'
      include 'TOTAL'
      include 'RESTART'

      logical if_full_pres_tmp

      real pm1(lx1*ly1*lz1,lelv)
      integer e

      nxyz2 = nx2*ny2*nz2

      if (ifmhd.and.ifile.eq.2) then
         do e=1,nelv
            if (if_full_pres) then
               call copy  (pm(1,1,1,e),pm1(1,e),nxyz2)
            else
               call map12 (pm(1,1,1,e),pm1(1,e),e)
            endif
         enddo
      elseif (ifpert.and.ifile.ge.2) then
         j=ifile-1     ! pointer to perturbation field
         if (ifsplit) then
            call copy (prp(1,j),pm1,nx1*ny1*nz1*nelv)
         else
            do e=1,nelv
               ie2 = (e-1)*nxyz2+1
               if (if_full_pres) then
                  call copy(prp(ie2,j),pm1(1,e),nxyz2)
               else
                  call map12(prp(ie2,j),pm1(1,e),e)
               endif
            enddo
         endif
      elseif (ifsplit) then
         call copy (pr,pm1,nx1*ny1*nz1*nelv)
      else
         do e=1,nelv
            if (if_full_pres) then
               call copy  (pr(1,1,1,e),pm1(1,e),nxyz2)
            else
               call map12 (pr(1,1,1,e),pm1(1,e),e)
            endif
         enddo
      endif
   
      return
      end
c-----------------------------------------------------------------------
c     To create file name; based on mfo_open_files from
      subroutine mfo_filename(prefix,fname,bname,k) 

      include 'SIZE'
      include 'INPUT'
      include 'PARALLEL'
      include 'RESTART'

      character*1 prefix(3)
      character*3 prefx

      character*132  fname, bname
      character*1   fnam1(132)

      integer k

      character*6  six
      save         six
      data         six / "??????" /

      character*1 slash,dot
      save        slash,dot
      data        slash,dot  / '/' , '.' /

      real rfileo

      call blank(fname,132)     !  zero out

      call chcopy(prefx,prefix,3)

#ifdef MPIIO
      rfileo = 1
#else
      rfileo = nfileo
#endif
      ndigit = log10(rfileo) + 1
     
      k = 1
      if (ifdiro) then          !  Add directory
         call chcopy(fnam1(1),'A',1)
         call chcopy(fnam1(2),six,ndigit)  ! put ???? in string
         k = 2 + ndigit
         call chcopy(fnam1(k),slash,1)
         k = k+1
      endif

      if (prefix(1).ne.' '.and.prefix(2).ne.' '.and. !  Add prefix
     $    prefix(3).ne.' ') then
         call chcopy(fnam1(k),prefix,3)
         k = k+3
      endif

      len=ltrunc(bname,132)   !  Add SESSION
      call chcopy(fnam1(k),bname,len)
      k = k+len
     
      if (ifreguo) then
         len=4
         call chcopy(fnam1(k),'_reg',len)
         k = k+len
      endif

      call chcopy(fnam1(k),six,ndigit) !  Add file-id holder
      k = k + ndigit

      call chcopy(fnam1(k  ),dot,1) !  Add .f appendix
      call chcopy(fnam1(k+1),'f',1)
      k = k + 2

      call chcopy(fname,fnam1,k)

      return
      end
c-----------------------------------------------------------------------
